@@ -1,10 +1,19 @@
 # Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 # SPDX-License-Identifier: Apache-2.0
+import sys
+from glob import glob
+from pathlib import Path
+
 import click
+import colorclass
+import luigi
 
 import pkg_resources
 import json
+
+import terminaltables
 from jinja2 import Template
+from luigi import LuigiStatusCode
 
 from servicecatalog_puppet import asset_helpers, manifest_utils, aws, luigi_tasks_and_targets
 from servicecatalog_puppet import constants
@@ -81,27 +90,18 @@ def write_share_template(portfolio_use_by_account, region, host_account_id, shar
         )
 
 
-def create_share_template(deployment_map, puppet_account_id):
+def create_share_template(deployment_map, import_map, puppet_account_id):
     logger.info("deployment_map: {}".format(deployment_map))
     ALL_REGIONS = get_regions()
     for region in ALL_REGIONS:
         logger.info("starting to build shares for region: {}".format(region))
         with betterboto_client.ClientContextManager('servicecatalog', region_name=region) as servicecatalog:
             portfolio_ids = {}
-            args = {}
-            while True:
 
-                response = servicecatalog.list_portfolios(
-                    **args
-                )
+            response = servicecatalog.list_portfolios_single_page()
 
-                for portfolio_detail in response.get('PortfolioDetails'):
-                    portfolio_ids[portfolio_detail.get('DisplayName')] = portfolio_detail.get('Id')
-
-                if response.get('PageToken') is not None:
-                    args['PageToken'] = response.get('PageToken')
-                else:
-                    break
+            for portfolio_detail in response.get('PortfolioDetails'):
+                portfolio_ids[portfolio_detail.get('DisplayName')] = portfolio_detail.get('Id')
 
             logger.info("Portfolios in use in region: {}".format(portfolio_ids))
 
@@ -110,10 +110,18 @@ def create_share_template(deployment_map, puppet_account_id):
                 if portfolio_use_by_account.get(account_id) is None:
                     portfolio_use_by_account[account_id] = []
                 for launch_id, launch in launch_details.get('launches').items():
-                    logger.info("portfolio ids: {}".format(portfolio_ids))
                     p = portfolio_ids[launch.get('portfolio')]
                     if p not in portfolio_use_by_account[account_id]:
                         portfolio_use_by_account[account_id].append(p)
+
+            for account_id, import_details in import_map.items():
+                if portfolio_use_by_account.get(account_id) is None:
+                    portfolio_use_by_account[account_id] = []
+                for spoke_local_portfolio_id, spoke_local_portfolio in import_details.get('spoke-local-portfolios').items():
+                    p = portfolio_ids[spoke_local_portfolio.get('portfolio')]
+                    if p not in portfolio_use_by_account[account_id]:
+                        portfolio_use_by_account[account_id].append(p)
+
             host_account_id = response.get('PortfolioDetails')[0].get('ARN').split(":")[4]
             sharing_policies = generate_bucket_policies_for_shares(deployment_map, puppet_account_id)
             write_share_template(portfolio_use_by_account, region, host_account_id, sharing_policies)
@@ -304,6 +312,12 @@ def _do_bootstrap_org_master(puppet_account_id, cloudformation, puppet_version):
                 'UsePreviousValue': False,
             },
         ],
+        'Tags':[
+            {
+                "Key": "ServiceCatalogPuppet:Actor",
+                "Value": "Framework",
+            }
+        ]
     }
     cloudformation.create_or_update(**args)
     response = cloudformation.describe_stacks(StackName=stack_name)
@@ -338,6 +352,12 @@ def _do_bootstrap_spoke(puppet_account_id, cloudformation, puppet_version):
                 'UsePreviousValue': False,
             },
         ],
+        'Tags':[
+            {
+                "Key": "ServiceCatalogPuppet:Actor",
+                "Value": "Framework",
+            }
+        ]
     }
     cloudformation.create_or_update(**args)
     logger.info('Finished bootstrap of spoke')
@@ -368,6 +388,12 @@ def _do_bootstrap(puppet_version):
                     'UsePreviousValue': False,
                 },
             ],
+            'Tags': [
+                {
+                    "Key": "ServiceCatalogPuppet:Actor",
+                    "Value": "Framework",
+                }
+            ]
         }
         for client_region, client in clients.items():
             process = Thread(name=client_region, target=client.create_or_update, kwargs=args)
@@ -474,17 +500,19 @@ def deploy_spoke_local_portfolios(manifest, launch_tasks):
                 )
                 tasks_to_run.append(import_into_spoke_local_portfolio_task)
 
-                create_launch_role_constraints_for_portfolio_task_params = {
-                    'launch_constraints': launch_details.get('constraints', {}).get('launch', []),
-                    'puppet_account_id': puppet_account_id,
-                }
-                create_launch_role_constraints_for_portfolio = luigi_tasks_and_targets.CreateLaunchRoleConstraintsForPortfolio(
-                    **create_spoke_local_portfolio_task_as_dependency_params,
-                    **import_into_spoke_local_portfolio_task_params,
-                    **create_launch_role_constraints_for_portfolio_task_params,
-                    dependencies=dependencies,
-                )
-                tasks_to_run.append(create_launch_role_constraints_for_portfolio)
+                launch_constraints = launch_details.get('constraints', {}).get('launch', [])
+                if len(launch_constraints) > 0:
+                    create_launch_role_constraints_for_portfolio_task_params = {
+                        'launch_constraints': launch_constraints,
+                        'puppet_account_id': puppet_account_id,
+                    }
+                    create_launch_role_constraints_for_portfolio = luigi_tasks_and_targets.CreateLaunchRoleConstraintsForPortfolio(
+                        **create_spoke_local_portfolio_task_as_dependency_params,
+                        **import_into_spoke_local_portfolio_task_params,
+                        **create_launch_role_constraints_for_portfolio_task_params,
+                        dependencies=dependencies,
+                    )
+                    tasks_to_run.append(create_launch_role_constraints_for_portfolio)
 
     return tasks_to_run
 
@@ -585,3 +613,112 @@ def deploy_launches_task_builder_for_account_launch_region(
 
     all_tasks[f"{task.get('account_id')}-{task.get('region')}-{task.get('launch_name')}"] = task
     return all_tasks
+
+
+def run_tasks(tasks_to_run):
+    for type in ["failure", "success", "timeout", "process_failure", "processing_time", "broken_task", ]:
+        os.makedirs(Path(constants.RESULTS_DIRECTORY) / type)
+
+    run_result = luigi.build(
+        tasks_to_run,
+        local_scheduler=True,
+        detailed_summary=True,
+        workers=10,
+        log_level='INFO',
+    )
+    table_data = [
+        ['Action','Launch', 'Account', 'Region', 'Portfolio', 'Product', 'Version', 'Duration'],
+
+    ]
+    table = terminaltables.AsciiTable(table_data)
+    for filename in glob('results/processing_time/*.json'):
+        result = json.loads(open(filename, 'r').read())
+        params = result.get('params_for_results')
+        table_data.append([
+            result.get('task_type'),
+            params.get('launch_name'),
+            params.get('account_id'),
+            params.get('region'),
+            params.get('portfolio'),
+            params.get('product'),
+            params.get('version'),
+            result.get('duration'),
+        ])
+    click.echo(table.table)
+    for filename in glob('results/failure/*.json'):
+        result = json.loads(open(filename, 'r').read())
+        click.echo(colorclass.Color("{red}" + result.get('task_type') + " failed{/red}"))
+        click.echo(f"{yaml.safe_dump({'parameters':result.get('task_params')})}")
+        click.echo("\n".join(result.get('exception_stack_trace')))
+        click.echo('')
+    exit_status_codes = {
+        LuigiStatusCode.SUCCESS: 0,
+        LuigiStatusCode.SUCCESS_WITH_RETRY: 0,
+        LuigiStatusCode.FAILED: 1,
+        LuigiStatusCode.FAILED_AND_SCHEDULING_FAILED: 2,
+        LuigiStatusCode.SCHEDULING_FAILED: 3,
+        LuigiStatusCode.NOT_RUN: 4,
+        LuigiStatusCode.MISSING_EXT: 5,
+    }
+    sys.exit(exit_status_codes.get(run_result.status))
+
+
+def run_tasks_for_dry_run(tasks_to_run):
+    for type in ["failure", "success", "timeout", "process_failure", "processing_time", "broken_task", ]:
+        os.makedirs(Path(constants.RESULTS_DIRECTORY) / type)
+
+    run_result = luigi.build(
+        tasks_to_run,
+        local_scheduler=True,
+        detailed_summary=True,
+        workers=10,
+        log_level='INFO',
+    )
+    for filename in glob('results/failure/*.json'):
+        result = json.loads(open(filename, 'r').read())
+        click.echo(colorclass.Color("{red}" + result.get('task_type') + " failed{/red}"))
+        click.echo(f"{yaml.safe_dump({'parameters':result.get('task_params')})}")
+        click.echo("\n".join(result.get('exception_stack_trace')))
+        click.echo('')
+    exit_status_codes = {
+        LuigiStatusCode.SUCCESS: 0,
+        LuigiStatusCode.SUCCESS_WITH_RETRY: 0,
+        LuigiStatusCode.FAILED: 1,
+        LuigiStatusCode.FAILED_AND_SCHEDULING_FAILED: 2,
+        LuigiStatusCode.SCHEDULING_FAILED: 3,
+        LuigiStatusCode.NOT_RUN: 4,
+        LuigiStatusCode.MISSING_EXT: 5,
+    }
+
+    click.echo("Dry run results")
+    table_data = [
+        ['Result','Launch', 'Account', 'Region', 'Current Version', 'New Version', 'Notes'],
+
+    ]
+    table = terminaltables.AsciiTable(table_data)
+    for filename in glob('output/TerminateProductDryRunTask/*.json'):
+        result = json.loads(open(filename, 'r').read())
+        table_data.append([
+            result.get('effect'),
+            result.get('params').get('launch_name'),
+            result.get('params').get('account_id'),
+            result.get('params').get('region'),
+            result.get('current_version'),
+            result.get('new_version'),
+            result.get('notes'),
+        ])
+    for filename in glob('output/ProvisionProductDryRunTask/*.json'):
+        result = json.loads(open(filename, 'r').read())
+        table_data.append([
+            result.get('effect'),
+            result.get('params').get('launch_name'),
+            result.get('params').get('account_id'),
+            result.get('params').get('region'),
+            result.get('current_version'),
+            result.get('new_version'),
+            result.get('notes'),
+        ])
+    click.echo(table.table)
+
+
+    sys.exit(exit_status_codes.get(run_result.status))
