@@ -1,18 +1,15 @@
 # Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 # SPDX-License-Identifier: Apache-2.0
-import sys
+from pathlib import Path
 
+import cfn_tools
+import requests
 from colorclass import Color
-from luigi import LuigiStatusCode
-from luigi.execution_summary import LuigiRunResult
-from terminaltables import AsciiTable
-
-import copy
+import terminaltables
 
 import shutil
 import json
 
-import luigi
 import pkg_resources
 import yaml
 import logging
@@ -23,10 +20,10 @@ from jinja2 import Template
 from pykwalify.core import Core
 from betterboto import client as betterboto_client
 
-
 from servicecatalog_puppet import cli_command_helpers
 from servicecatalog_puppet import luigi_tasks_and_targets
 from servicecatalog_puppet import manifest_utils
+from servicecatalog_puppet import aws
 
 
 from servicecatalog_puppet import asset_helpers
@@ -54,7 +51,33 @@ def generate_shares(f):
 
     manifest = manifest_utils.load(f)
     deployment_map = manifest_utils.build_deployment_map(manifest, constants.LAUNCHES)
-    cli_command_helpers.create_share_template(deployment_map, cli_command_helpers.get_puppet_account_id())
+    import_map = manifest_utils.build_deployment_map(manifest, constants.SPOKE_LOCAL_PORTFOLIOS)
+    cli_command_helpers.create_share_template(deployment_map, import_map, cli_command_helpers.get_puppet_account_id())
+
+
+def dry_run(f):
+    manifest = manifest_utils.load(f)
+
+    launch_tasks = {}
+    tasks_to_run = []
+
+    all_launch_tasks = cli_command_helpers.deploy_launches(manifest)
+    launch_tasks.update(all_launch_tasks)
+
+    for task in cli_command_helpers.wire_dependencies(launch_tasks):
+        task_status = task.get('status')
+        del task['status']
+        if task_status == constants.PROVISIONED:
+            tasks_to_run.append(luigi_tasks_and_targets.ProvisionProductDryRunTask(**task))
+        elif task_status == constants.TERMINATED:
+            tasks_to_run.append(luigi_tasks_and_targets.TerminateProductDryRunTask(**task))
+        else:
+            raise Exception(f"Unsupported status of {task_status}")
+
+    # spoke_local_portfolio_tasks_to_run = cli_command_helpers.deploy_spoke_local_portfolios(manifest, launch_tasks)
+    # tasks_to_run += spoke_local_portfolio_tasks_to_run
+
+    cli_command_helpers.run_tasks_for_dry_run(tasks_to_run)
 
 
 def deploy(f, single_account):
@@ -85,9 +108,6 @@ def deploy(f, single_account):
                     else:
                         raise Exception(f"Launch {task.get('launch_name')} has disallowed attribute: {attribute}")
 
-            for a in ['parameters', 'ssm_param_inputs', 'outputs', 'dependencies']:
-                if task.get(a, None) is not None:
-                    del task[a]
             tasks_to_run.append(luigi_tasks_and_targets.TerminateProductTask(**task))
         else:
             raise Exception(f"Unsupported status of {task_status}")
@@ -95,25 +115,7 @@ def deploy(f, single_account):
     spoke_local_portfolio_tasks_to_run = cli_command_helpers.deploy_spoke_local_portfolios(manifest, launch_tasks)
     tasks_to_run += spoke_local_portfolio_tasks_to_run
 
-    result = luigi.build(
-        tasks_to_run,
-        local_scheduler=True,
-        detailed_summary=True,
-        workers=10,
-        log_level='INFO',
-    )
-
-    # click.echo(f"Workflow complete {result} - {result.status.value[1]}")
-    exit_status_codes = {
-        LuigiStatusCode.SUCCESS: 0,
-        LuigiStatusCode.SUCCESS_WITH_RETRY: 0,
-        LuigiStatusCode.FAILED: 1,
-        LuigiStatusCode.FAILED_AND_SCHEDULING_FAILED: 2,
-        LuigiStatusCode.SCHEDULING_FAILED:3,
-        LuigiStatusCode.NOT_RUN:4,
-        LuigiStatusCode.MISSING_EXT:5,
-    }
-    sys.exit(exit_status_codes.get(result.status))
+    cli_command_helpers.run_tasks(tasks_to_run)
 
 
 def bootstrap_spoke_as(puppet_account_id, iam_role_arns):
@@ -259,7 +261,7 @@ def list_launches(f):
                         active,
                         status,
                     ])
-    click.echo(AsciiTable(table).table)
+    click.echo(terminaltables.AsciiTable(table).table)
 
 
 def expand(f):
@@ -411,3 +413,73 @@ def quick_start():
     click.echo("All done!")
 
 
+def run(what, tail):
+    pipelines = {
+        'puppet': constants.PIPELINE_NAME
+    }
+    pipeline_name = pipelines.get(what)
+    pipeline_execution_id = aws.run_pipeline(pipeline_name, tail)
+    click.echo(
+        f"https://{os.environ.get('AWS_DEFAULT_REGION')}.console.aws.amazon.com/codesuite/codepipeline/pipelines/{pipeline_name}/executions/{pipeline_execution_id}/timeline"
+    )
+
+
+def list_resources():
+    click.echo("# Framework resources")
+
+    click.echo("## SSM Parameters used")
+    click.echo(f"- {constants.CONFIG_PARAM_NAME}")
+    click.echo(f"- {constants.CONFIG_PARAM_NAME_ORG_IAM_ROLE_ARN}")
+
+    for file in Path(__file__).parent.resolve().glob("*.template.yaml"):
+        if 'empty.template.yaml' == file.name:
+            continue
+        template_contents = Template(open(file, 'r').read()).render()
+        template = cfn_tools.load_yaml(template_contents)
+        click.echo(f"## Resources for stack: {file.name.split('.')[0]}")
+        table_data = [
+            ['Logical Name', 'Resource Type', 'Name', ],
+        ]
+        table = terminaltables.AsciiTable(table_data)
+        for logical_name, resource in template.get('Resources').items():
+            resource_type = resource.get('Type')
+            name = '-'
+            type_to_name = {
+                'AWS::IAM::Role': 'RoleName',
+                'AWS::SSM::Parameter': 'Name',
+                'AWS::S3::Bucket': 'BucketName',
+                'AWS::CodePipeline::Pipeline': 'Name',
+                'AWS::CodeBuild::Project': 'Name',
+                'AWS::CodeCommit::Repository': 'RepositoryName',
+                'AWS::SNS::Topic': 'TopicName',
+                'AWS::SQS::Queue': 'QueueName',
+            }
+
+            if type_to_name.get(resource_type) is not None:
+                name = resource.get('Properties', {}).get(type_to_name.get(resource_type), 'Not Specified')
+                if not isinstance(name, str):
+                    name = cfn_tools.dump_yaml(name)
+
+            table_data.append([logical_name, resource_type, name])
+
+        click.echo(table.table)
+    click.echo(f"n.b. AWS::StackName evaluates to {constants.BOOTSTRAP_STACK_NAME}")
+
+
+def import_product_set(f, name, portfolio_name):
+    url = f"https://raw.githubusercontent.com/awslabs/aws-service-catalog-products/master/{name}/manifest.yaml"
+    response = requests.get(url)
+    logger.info(
+        f"Getting {url}"
+    )
+    manifest = yaml.safe_load(f.read())
+    if manifest.get('launches') is None:
+        manifest['launches'] = {}
+    manifest_segment = yaml.safe_load(response.text)
+    for launch_name, details in manifest_segment.get('launches').items():
+        details['portfolio'] = portfolio_name
+        manifest['launches'][launch_name] = details
+    with open(f.name, 'w') as f:
+        f.write(
+            yaml.safe_dump(manifest)
+        )
