@@ -1,8 +1,12 @@
+import copy
+import re
 import time
+import traceback
+from pathlib import Path
 
 from betterboto import client as betterboto_client
 
-from servicecatalog_puppet import aws, cli_command_helpers
+from servicecatalog_puppet import aws, cli_command_helpers, constants
 
 import luigi
 import json
@@ -37,11 +41,33 @@ class SSMParamTarget(luigi.Target):
             ).get('Parameter').get('Value')
 
 
-class GetSSMParamTask(luigi.Task):
+class PuppetTask(luigi.Task):
+
+    @property
+    def resources(self):
+        resources_for_this_task = {}
+
+        if hasattr(self, 'region'):
+            resources_for_this_task[self.region] = 1
+
+        return resources_for_this_task
+
+    def params_for_results_display(self):
+        return "Omitted"
+
+
+class GetSSMParamTask(PuppetTask):
     parameter_name = luigi.Parameter()
     name = luigi.Parameter()
     region = luigi.Parameter(default=None)
 
+    def params_for_results_display(self):
+        return {
+            "parameter_name": self.parameter_name,
+            "name": self.name,
+            "region": self.region,
+        }
+
     def output(self):
         return SSMParamTarget(self.name, True)
 
@@ -49,7 +75,7 @@ class GetSSMParamTask(luigi.Task):
         pass
 
 
-class ProvisionProductTask(luigi.Task):
+class ProvisionProductTask(PuppetTask):
     launch_name = luigi.Parameter()
     portfolio = luigi.Parameter()
     product = luigi.Parameter()
@@ -68,17 +94,12 @@ class ProvisionProductTask(luigi.Task):
 
     retry_count = luigi.IntParameter(default=1)
 
-    status = luigi.Parameter(default='', significant=False)
-
     worker_timeout = luigi.IntParameter(default=0, significant=False)
 
     ssm_param_outputs = luigi.ListParameter(default=[])
 
     try_count = 1
 
-    def add_requires(self, task):
-        self.reqs.append(task)
-
     def requires(self):
         ssm_params = {}
         for param_input in self.ssm_param_inputs:
@@ -91,10 +112,20 @@ class ProvisionProductTask(luigi.Task):
             'ssm_params': ssm_params
         }
 
+    def params_for_results_display(self):
+        return {
+            "launch_name": self.launch_name,
+            "account_id": self.account_id,
+            "region": self.region,
+            "portfolio": self.portfolio,
+            "product": self.product,
+            "version": self.version,
+        }
+
     def output(self):
         return luigi.LocalTarget(
             f"output/ProvisionProductTask/"
-            f"{self.account_id}-{self.region}-{self.portfolio}-{self.product}-{self.version}.json"
+            f"{self.launch_name}-{self.account_id}-{self.region}-{self.portfolio}-{self.product}-{self.version}.json"
         )
 
     def run(self):
@@ -232,7 +263,203 @@ class ProvisionProductTask(luigi.Task):
                 logger.info(f"[{self.launch_name}] {self.account_id}:{self.region} :: finished provisioning")
 
 
-class TerminateProductTask(luigi.Task):
+class ProvisionProductDryRunTask(PuppetTask):
+    launch_name = luigi.Parameter()
+    portfolio = luigi.Parameter()
+    product = luigi.Parameter()
+    version = luigi.Parameter()
+
+    product_id = luigi.Parameter()
+    version_id = luigi.Parameter()
+
+    account_id = luigi.Parameter()
+    region = luigi.Parameter()
+    puppet_account_id = luigi.Parameter()
+
+    parameters = luigi.ListParameter(default=[])
+    ssm_param_inputs = luigi.ListParameter(default=[])
+    dependencies = luigi.ListParameter(default=[])
+
+    retry_count = luigi.IntParameter(default=1)
+
+    worker_timeout = luigi.IntParameter(default=0, significant=False)
+
+    ssm_param_outputs = luigi.ListParameter(default=[])
+
+    try_count = 1
+
+    def requires(self):
+        dependencies = []
+        for r in self.dependencies:
+            # if hasattr(r, 'status'):
+            #     del r['status']
+            dependencies.append(
+                self.__class__(**r)
+            )
+
+        return {
+            'dependencies': dependencies,
+        }
+
+    def params_for_results_display(self):
+        return {
+            "launch_name": self.launch_name,
+            "account_id": self.account_id,
+            "region": self.region,
+            "portfolio": self.portfolio,
+            "product": self.product,
+            "version": self.version,
+        }
+
+    def output(self):
+        return luigi.LocalTarget(
+            f"output/ProvisionProductDryRunTask/"
+            f"{self.launch_name}-{self.account_id}-{self.region}-{self.portfolio}-{self.product}-{self.version}.json"
+        )
+
+    def run(self):
+        logger_prefix = f"[{self.launch_name}] {self.account_id}:{self.region}"
+        logger.info(
+            f"{logger_prefix} :: starting dryrun of {self.retry_count}"
+        )
+
+        role = f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole"
+        with betterboto_client.CrossAccountClientContextManager(
+                'servicecatalog', role, f'sc-{self.region}-{self.account_id}', region_name=self.region
+        ) as service_catalog:
+            logger.info(f"{logger_prefix} :: looking for previous failures")
+
+            response = service_catalog.search_provisioned_products(
+                Filters={'SearchQuery': [
+                    "productId:{}".format(self.product_id)
+                ]}
+            )
+            provisioning_artifact_id = False
+            for r in response.get('ProvisionedProducts', []):
+                if r.get('Name') == self.launch_name:
+                    current_status = r.get('Status')
+                    if current_status in ["AVAILABLE", "TAINTED"]:
+                        provisioned_product_id = r.get('Id')
+                        provisioning_artifact_id = r.get('ProvisioningArtifactId')
+
+                        if provisioning_artifact_id != self.version_id:
+                            self.write_result(
+                                current_version=self.get_current_version(provisioning_artifact_id),
+                                new_version=self.version,
+                                effect=constants.CHANGE,
+                                notes='Version change',
+                            )
+                            return
+                        else:
+                            all_params = {}
+                            logger.info(f"{logger_prefix} :: collecting ssm params")
+                            for ssm_param_name, ssm_param in self.input().get('ssm_params', {}).items():
+                                all_params[ssm_param_name] = ssm_param.read()
+                            logger.info(f"{logger_prefix} :: collecting manifest params")
+                            for parameter in self.parameters:
+                                all_params[parameter.get('name')] = parameter.get('value')
+
+                            with betterboto_client.CrossAccountClientContextManager(
+                                'cloudformation', role, f'cfn-{self.region}-{self.account_id}', region_name=self.region
+                            ) as cloudformation:
+                                default_cfn_params = aws.get_default_parameters_for_stack(
+                                  cloudformation, f"SC-{self.account_id}-{provisioned_product_id}"
+                              )
+
+                            for default_cfn_param_name in default_cfn_params.keys():
+                                if all_params.get(default_cfn_param_name) is None:
+                                    all_params[default_cfn_param_name] = default_cfn_params[default_cfn_param_name]
+
+                            logger.info(
+                                f"{logger_prefix} :: found previous good provision"
+                            )
+                            if provisioned_product_id:
+                                logger.info(
+                                    f"{logger_prefix} :: checking params for diffs"
+                                )
+                                provisioned_parameters = aws.get_parameters_for_stack(
+                                    cloudformation,
+                                    f"SC-{self.account_id}-{provisioned_product_id}"
+                                )
+                                logger.info(f"{logger_prefix} :: "
+                                            f"current params: {provisioned_parameters}")
+
+                                logger.info(f"{logger_prefix} :: "
+                                            f"new params: {all_params}")
+
+                                if provisioned_parameters == all_params:
+                                    logger.info(f"{logger_prefix} :: params unchanged")
+                                    self.write_result(
+                                        current_version=self.version,
+                                        new_version=self.version,
+                                        effect=constants.NO_CHANGE,
+                                        notes="Versions and params are the same",
+                                    )
+                                    return
+                                else:
+                                    logger.info(f"{logger_prefix} :: params changed")
+                                    self.write_result(
+                                        current_version=self.version,
+                                        new_version=self.version,
+                                        effect=constants.CHANGE,
+                                        notes="Versions are the same but the params are different",
+                                    )
+                                    return
+
+                    else:
+                        current_version = self.get_current_version(provisioning_artifact_id, service_catalog)
+
+                        if current_status in ["UNDER_CHANGE", "PLAN_IN_PROGRESS"]:
+                            self.write_result(
+                                current_version=current_version,
+                                new_version=current_version,
+                                effect=constants.NO_CHANGE,
+                                notes=f"provisioned product status was {current_status}"
+                            )
+                            return
+
+                        elif current_status == 'ERROR':
+                            self.write_result(
+                                current_version=current_version,
+                                new_version=self.version,
+                                effect=constants.CHANGE,
+                                notes=f"provisioned product status was {current_status}"
+                            )
+                            return
+
+            logger.info(f"{logger_prefix} no version of product has not been provisioned")
+            self.write_result(
+                current_version=None,
+                effect=constants.CHANGE,
+                new_version=self.version,
+                notes="Fresh install",
+            )
+
+    def get_current_version(self, provisioning_artifact_id, service_catalog):
+        return service_catalog.describe_provisioning_artifact(
+            ProvisioningArtifactId=provisioning_artifact_id,
+            ProductId=self.product_id,
+        ).get('ProvisioningArtifactDetail').get('Name')
+
+    def write_result(self, current_version, new_version, effect, notes=''):
+        f = self.output().open('w')
+        f.write(
+            json.dumps(
+                {
+                    "current_version": current_version,
+                    "new_version": new_version,
+                    "effect": effect,
+                    "notes": notes,
+                    "params": self.param_kwargs
+                },
+                indent=4,
+                default=str,
+            )
+        )
+        f.close()
+
+
+class TerminateProductTask(PuppetTask):
     launch_name = luigi.Parameter()
     portfolio = luigi.Parameter()
     product = luigi.Parameter()
@@ -247,10 +474,26 @@ class TerminateProductTask(luigi.Task):
 
     retry_count = luigi.IntParameter(default=1)
 
+    ssm_param_outputs = luigi.ListParameter(default=[])
+
     worker_timeout = luigi.IntParameter(default=0, significant=False)
 
     try_count = 1
 
+    parameters = luigi.ListParameter(default=[])
+    ssm_param_inputs = luigi.ListParameter(default=[])
+    dependencies = luigi.ListParameter(default=[])
+
+
+    def params_for_results_display(self):
+        return {
+            "account_id": self.account_id,
+            "region": self.region,
+            "portfolio": self.portfolio,
+            "product": self.product,
+            "version": self.version,
+        }
+
     def output(self):
         return luigi.LocalTarget(
             f"output/TerminateProductTask/"
@@ -273,6 +516,25 @@ class TerminateProductTask(luigi.Task):
             log_output.update({
                 "provisioned_product_id": provisioned_product_id,
             })
+
+            for ssm_param_output in self.ssm_param_outputs:
+                param_name = ssm_param_output.get('param_name')
+                logger.info(
+                    f"[{self.launch_name}] {self.account_id}:{self.region} :: deleting SSM Param: {param_name}"
+                )
+                with betterboto_client.ClientContextManager('ssm') as ssm:
+                    try:
+                        ssm.delete_parameter(
+                            Name=param_name,
+                        )
+                        logger.info(
+                            f"[{self.launch_name}] {self.account_id}:{self.region} :: deleting SSM Param: {param_name}"
+                        )
+                    except ssm.exceptions.ParameterNotFound:
+                        logger.info(
+                            f"[{self.launch_name}] {self.account_id}:{self.region} :: SSM Param: {param_name} not found"
+                        )
+
             f = self.output().open('w')
             f.write(
                 json.dumps(
@@ -285,7 +547,97 @@ class TerminateProductTask(luigi.Task):
             logger.info(f"[{self.launch_name}] {self.account_id}:{self.region} :: finished terminating")
 
 
-class CreateSpokeLocalPortfolioTask(luigi.Task):
+class TerminateProductDryRunTask(PuppetTask):
+    launch_name = luigi.Parameter()
+    portfolio = luigi.Parameter()
+    product = luigi.Parameter()
+    version = luigi.Parameter()
+
+    product_id = luigi.Parameter()
+    version_id = luigi.Parameter()
+
+    account_id = luigi.Parameter()
+    region = luigi.Parameter()
+    puppet_account_id = luigi.Parameter()
+
+    retry_count = luigi.IntParameter(default=1)
+
+    ssm_param_outputs = luigi.ListParameter(default=[])
+
+    worker_timeout = luigi.IntParameter(default=0, significant=False)
+
+    try_count = 1
+
+    parameters = luigi.ListParameter(default=[])
+    ssm_param_inputs = luigi.ListParameter(default=[])
+    dependencies = luigi.ListParameter(default=[])
+
+
+    def params_for_results_display(self):
+        return {
+            "launch_name": self.launch_name,
+            "account_id": self.account_id,
+            "region": self.region,
+            "portfolio": self.portfolio,
+            "product": self.product,
+            "version": self.version,
+        }
+
+    def output(self):
+        return luigi.LocalTarget(
+            f"output/TerminateProductDryRunTask/"
+            f"{self.launch_name}-{self.account_id}-{self.region}-{self.portfolio}-{self.product}-{self.version}.json"
+        )
+
+    def write_result(self, current_version, new_version, effect, notes=''):
+        f = self.output().open('w')
+        f.write(
+            json.dumps(
+                {
+                    "current_version": current_version,
+                    "new_version": new_version,
+                    "effect": effect,
+                    "notes": notes,
+                    "params": self.param_kwargs
+                },
+                indent=4,
+                default=str,
+            )
+        )
+        f.close()
+
+    def run(self):
+        logger.info(f"[{self.launch_name}] {self.account_id}:{self.region} :: "
+                    f"starting dry run terminate try {self.try_count} of {self.retry_count}")
+
+        role = f"arn:aws:iam::{self.account_id}:role/servicecatalog-puppet/PuppetRole"
+        with betterboto_client.CrossAccountClientContextManager(
+                'servicecatalog', role, f'sc-{self.region}-{self.account_id}', region_name=self.region
+        ) as service_catalog:
+            logger.info(f"[{self.launch_name}] {self.account_id}:{self.region} :: looking for previous failures")
+            r = aws.get_provisioned_product_details(self.product_id, self.launch_name, service_catalog)
+
+            if r is None:
+                self.write_result(
+                    '-', '-', constants.NO_CHANGE, notes='There is nothing to terminate'
+                )
+            else:
+                provisioned_product_name = service_catalog.describe_provisioning_artifact(
+                    ProvisioningArtifactId=r.get('ProvisioningArtifactId'),
+                    ProductId=self.product_id,
+                ).get('ProvisioningArtifactDetail').get('Name')
+
+                if r.get('Status') != "TERMINATED":
+                    self.write_result(
+                        provisioned_product_name, '-', constants.CHANGE, notes='The product would be terminated'
+                    )
+                else:
+                    self.write_result(
+                        '-', '-', constants.CHANGE, notes='The product is already terminated'
+                    )
+
+
+class CreateSpokeLocalPortfolioTask(PuppetTask):
     account_id = luigi.Parameter()
     region = luigi.Parameter()
     portfolio = luigi.Parameter()
@@ -293,6 +645,13 @@ class CreateSpokeLocalPortfolioTask(luigi.Task):
     provider_name = luigi.Parameter(significant=False, default='not set')
     description = luigi.Parameter(significant=False, default='not set')
 
+    def params_for_results_display(self):
+        return {
+            "account_id": self.account_id,
+            "region": self.region,
+            "portfolio": self.portfolio,
+        }
+
     def output(self):
         return luigi.LocalTarget(
             f"output/CreateSpokeLocalPortfolioTask/"
@@ -323,7 +682,7 @@ class CreateSpokeLocalPortfolioTask(luigi.Task):
         logger.info(f"[{self.portfolio}] {self.account_id}:{self.region} :: finished creating portfolio")
 
 
-class CreateAssociationsForPortfolioTask(luigi.Task):
+class CreateAssociationsForPortfolioTask(PuppetTask):
     account_id = luigi.Parameter()
     region = luigi.Parameter()
     portfolio = luigi.Parameter()
@@ -342,6 +701,13 @@ class CreateAssociationsForPortfolioTask(luigi.Task):
             'deps': [ProvisionProductTask(**dependency) for dependency in self.dependencies]
         }
 
+    def params_for_results_display(self):
+        return {
+            "account_id": self.account_id,
+            "region": self.region,
+            "portfolio": self.portfolio,
+        }
+
     def output(self):
         return luigi.LocalTarget(
             f"output/CreateAssociationsForPortfolioTask/"
@@ -388,7 +754,7 @@ class CreateAssociationsForPortfolioTask(luigi.Task):
             logger.info(f"[{self.portfolio}] {self.account_id}:{self.region} :: Finished importing")
 
 
-class ImportIntoSpokeLocalPortfolioTask(luigi.Task):
+class ImportIntoSpokeLocalPortfolioTask(PuppetTask):
     account_id = luigi.Parameter()
     region = luigi.Parameter()
     portfolio = luigi.Parameter()
@@ -402,6 +768,14 @@ class ImportIntoSpokeLocalPortfolioTask(luigi.Task):
             portfolio=self.portfolio,
         )
 
+    def params_for_results_display(self):
+        return {
+            "account_id": self.account_id,
+            "region": self.region,
+            "portfolio": self.portfolio,
+            "hub_portfolio_id": self.hub_portfolio_id,
+        }
+
     def output(self):
         return luigi.LocalTarget(
             f"output/ImportIntoSpokeLocalPortfolioTask/"
@@ -546,7 +920,7 @@ class ImportIntoSpokeLocalPortfolioTask(luigi.Task):
         logger.info(f"[{self.portfolio}] {self.account_id}:{self.region} :: Finished importing")
 
 
-class CreateLaunchRoleConstraintsForPortfolio(luigi.Task):
+class CreateLaunchRoleConstraintsForPortfolio(PuppetTask):
     account_id = luigi.Parameter()
     region = luigi.Parameter()
     portfolio = luigi.Parameter()
@@ -579,15 +953,41 @@ class CreateLaunchRoleConstraintsForPortfolio(luigi.Task):
         with betterboto_client.CrossAccountClientContextManager(
                 'cloudformation', role, f'cfn-{self.account_id}-{self.region}', region_name=self.region
         ) as cloudformation:
+            new_launch_constraints = []
+            for launch_constraint in self.launch_constraints:
+                new_launch_constraint = {
+                    'products': [],
+                    'roles': launch_constraint.get('roles')
+                }
+                if launch_constraint.get('products', None) is not None:
+                    if isinstance(launch_constraint.get('products'), tuple):
+                        new_launch_constraint['products'] += launch_constraint.get('products')
+                    elif isinstance(launch_constraint.get('products'), str):
+                        with betterboto_client.CrossAccountClientContextManager(
+                                'servicecatalog', role, f'sc-{self.account_id}-{self.region}', region_name=self.region
+                        ) as service_catalog:
+                            response = service_catalog.search_products_as_admin_single_page(PortfolioId=portfolio_id)
+                            for product_view_details in response.get('ProductViewDetails', []):
+                                product_view_summary = product_view_details.get('ProductViewSummary')
+                                product_name_to_id_dict[product_view_summary.get('Name')] = product_view_summary.get(
+                                    'ProductId')
+                                if re.match(launch_constraint.get('products'), product_view_summary.get('Name')):
+                                    new_launch_constraint['products'].append(product_view_summary.get('Name'))
+
+                if launch_constraint.get('product', None) is not None:
+                    new_launch_constraint['products'].append(launch_constraint.get('product'))
+
+                new_launch_constraints.append(new_launch_constraint)
+
             template = cli_command_helpers.env.get_template('launch_role_constraints.template.yaml.j2').render(
                 portfolio={
                     'DisplayName': self.portfolio,
                 },
                 portfolio_id=portfolio_id,
-                launch_constraints=self.launch_constraints,
+                launch_constraints=new_launch_constraints,
                 product_name_to_id_dict=product_name_to_id_dict,
             )
-            time.sleep(60)
+            time.sleep(30)
             stack_name = f"launch-constraints-for-portfolio-{portfolio_id}"
             cloudformation.create_or_update(
                 StackName=stack_name,
@@ -609,8 +1009,79 @@ class CreateLaunchRoleConstraintsForPortfolio(luigi.Task):
             )
             f.close()
 
+    def params_for_results_display(self):
+        return {
+            "account_id": self.account_id,
+            "region": self.region,
+            "portfolio": self.portfolio,
+            "hub_portfolio_id": self.hub_portfolio_id,
+        }
+
     def output(self):
         return luigi.LocalTarget(
             f"output/CreateLaunchRoleConstraintsForPortfolio/"
             f"{self.account_id}-{self.region}-{self.portfolio}-{self.hub_portfolio_id}.json"
         )
+
+
+def record_event(event_type, task, extra_event_data=None):
+    task_type = task.__class__.__name__
+    task_params = task.param_kwargs
+
+    event = {
+        "event_type": event_type,
+        "task_type": task_type,
+        "task_params": task_params,
+        "params_for_results": task.params_for_results_display(),
+    }
+    if extra_event_data is not None:
+        event.update(extra_event_data)
+
+    with open(
+            Path(constants.RESULTS_DIRECTORY) / event_type / f"{task_type}-{task.task_id}.json", 'w'
+    ) as f:
+        f.write(
+            json.dumps(
+                event,
+                default=str,
+                indent=4,
+            )
+        )
+
+
+@luigi.Task.event_handler(luigi.Event.FAILURE)
+def on_task_failure(task, exception):
+    exception_details = {
+        "exception_type": type(exception),
+        "exception_stack_trace": traceback.format_exception(
+            etype=type(exception),
+            value=exception,
+            tb=exception.__traceback__,
+        )
+    }
+    record_event('failure', task, exception_details)
+
+
+@luigi.Task.event_handler(luigi.Event.SUCCESS)
+def on_task_success(task):
+    record_event('success', task)
+
+
+@luigi.Task.event_handler(luigi.Event.TIMEOUT)
+def on_task_timeout(task):
+    record_event('timeout', task)
+
+
+@luigi.Task.event_handler(luigi.Event.PROCESS_FAILURE)
+def on_task_process_failure(task):
+    record_event('process_failure', task)
+
+
+@luigi.Task.event_handler(luigi.Event.PROCESSING_TIME)
+def on_task_processing_time(task, duration):
+    record_event('processing_time', task, {"duration": duration})
+
+
+@luigi.Task.event_handler(luigi.Event.BROKEN_TASK)
+def on_task_broken_task(task):
+    record_event('broken_task', task)
